module Array where

  open import Prelude

  Data : Set
  Data = Nat

  Array : Nat -> Set
  Array n = Fin n -> Data

  Shape : Set
  Shape = List Nat

  data Heap : Shape -> Set where
    Empty : Heap Nil
    Alloc : {n : Nat} -> {ns : Shape} -> 
      Array n -> Heap ns -> Heap (Cons n ns)

  data Loc  : Nat -> Shape -> Set where
    Top  : {n : Nat} -> {ns : Shape} -> Loc n (Cons n ns)
    Pop  : forall {n k ns} -> Loc n ns -> Loc n (Cons k ns)

  data IO (a : Set) : Shape -> Shape -> Set where
    Return : {ns : Shape} -> a -> IO a ns ns  
    Write : forall {n ns ms} -> 
      Loc n ns -> Fin n -> Data -> IO a ns ms -> IO a ns ms
    Read : forall {n ns ms} ->
      Loc n ns -> Fin n -> (Data -> IO a ns ms) -> IO a ns ms
    New : forall {ns ms} -> 
      (n : Nat) -> (Loc n (Cons n ns) -> IO a (Cons n ns) ms) -> 
      IO a ns ms

  return : {a : Set} -> {ns : Shape} -> a -> IO a ns ns
  return x = Return x

  _>>=_ : forall {a b ns ms ks} ->
    IO a ns ms -> (a -> IO b ms ks) -> IO b ns ks
  Return x >>= f        = f x
  Write a i x wr >>= f  = Write a i x (wr >>= f)
  Read a i rd >>= f     = Read a i (\x -> rd x >>= f)
  New n io >>= f        = New n (\a -> io a >>= f)

  _>>_ : {ns ms ks : Shape} {A B : Set} -> 
    IO A ns ms -> IO B ms ks -> IO B ns ks
  m >> n = m >>= \i -> n

  inj : forall {ms ns n} -> Loc n ns -> Loc n (append ms ns)
  inj  {Nil}         i      = i
  inj  {Cons k ks}   i      = Pop (inj {ks} i)

  lookup : forall {n ns} -> Loc n ns -> Fin n -> Heap ns -> Data
  lookup  Top      i  (Alloc a _)  = a i
  lookup  (Pop k)  i  (Alloc _ h)  = lookup k i h

  updateArray : {n : Nat} -> Fin n -> Data -> Array n -> Array n
  updateArray i d a = \j -> if i == j then d else a j

  updateHeap : forall {n ns} -> 
    Loc n ns -> Fin n -> Data -> Heap ns -> Heap ns
  updateHeap Top      i x (Alloc a h)  = Alloc (updateArray i x a) h
  updateHeap (Pop k)  i x (Alloc a h)  = Alloc a (updateHeap k i x h)

  run : forall {a ns ms} -> IO a ns ms -> Heap ns -> Pair a (Heap ms)
  run (Return x) h         = pair x h
  run (Read a i rd) h      = run (rd (lookup a i h)) h
  run (Write a i x wr) h   = run wr (updateHeap a i x h)
  run (New n io) h         = run (io Top) (Alloc (\i -> Zero) h)
