module Prelude where

  data Bool : Set where
    True : Bool
    False : Bool

  if_then_else_ : {A : Set} -> Bool -> A -> A -> A
  if True then t else e = t
  if False then t else e = e

  data Unit : Set where
    void : Unit

  data Pair (a : Set) (b : Set) : Set where
    pair : a -> b -> Pair a b

  fst : forall {a b} -> Pair a b -> a
  fst (pair x y) = x

  snd : forall {a b} -> Pair a b -> b
  snd (pair x y) = y


  data Nat : Set where
    Zero   : Nat
    Succ   : Nat -> Nat

  _+_ : Nat -> Nat -> Nat
  Zero    + m  = m
  Succ n  + m  = Succ (n + m)

  {-#  BUILTIN NATURAL Nat  #-}
  {-#  BUILTIN SUC Succ  #-}
  {-#  BUILTIN ZERO Zero  #-}  
  {-#  BUILTIN NATPLUS _+_  #-}  
 
  _*_ : Nat -> Nat -> Nat
  Zero * m = Zero
  (Succ k) * m = m + (k * m)

  data List (a : Set) : Set where
    Nil : List a  
    Cons : a -> List a -> List a

  append : {a : Set} -> List a -> List a -> List a
  append  Nil          ys  = ys 
  append  (Cons x xs)  ys  = Cons x (append xs ys) 

  data Fin : Nat -> Set where
    Fz  : {n : Nat} -> Fin (Succ n)
    Fs  : {n : Nat} -> Fin n -> Fin (Succ n)

  _==_ : {n m : Nat} -> Fin n -> Fin m -> Bool
  Fz == Fz = True
  Fs k == Fs j = k == j
  Fz == Fs k = False
  Fs k == Fz = False

  fmax : {n : Nat} -> Fin (Succ n)
  fmax {Zero} = Fz
  fmax {Succ k} = Fs fmax

  data _===_ {A : Set} : A -> A -> Set where
    Refl : {x : A} -> x === x

  _o_ : {A B C : Set} -> (B -> C) -> (A -> B) -> (A -> C)
  g o f = \x -> g (f x) 


